//
//   File: makewave.c      Created 10-DEC-2001  Richard B. Johnson
//                                              rjohnson@analogic.com
//   Released under GPL.
//
//   This makes a ".wav" file containing left-channel and right-channel
//   tones that you can specify. By default, it produces a 10 second
//   file with 'C' as the left channel and 'A' as the right channel,
//   modulated 100 percent. This can be used to test the levels and
//   distortion in a CDROM record setup. The harmonic distortion of
//   both tones at frequencies under 10 kHz is well below 0.001 percent.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>


#ifdef _MSC_VER
#include "stdint.h"
#include <io.h>
#else
#include <unistd.h>
#include <stdint.h>
#include <endian.h>
#define O_BINARY 0
#endif

#include "wav.h"


#define SAMPLE_RATE 44100
#define NR_CHANS    2
#define SECONDS     10 
#define LEFT_FREQ   261.63 
#define RIGHT_FREQ  440.0
#define FP_ROUND    (double)0.499999 
#if !defined(PI)
# define PI 3.1415926535897932384626433832795029
#endif

static const char rif[]="RIFF";
static const char dat[]="data";
static const char fmt[]="WAVEfmt ";
static const char nam[]="untitled.wav";
static WAV wav;

#define FAIL -1
#define ERRORS(s) \
{                 \
    fprintf(stderr, "Error on line %d, file %s, call "#s", (%s)\n", \
    __LINE__, __FILE__, strerror(errno)); \
    exit(EXIT_FAILURE);  \
}

int main(int args, char *argv[])
{
    int fd;
    size_t i, lim, len, sec;
    int16_t  *sp, *buf, samp;
    const char *fname;
    double theta, twopi, left_freq,
    right_freq, left_mag, right_mag,
    dac_max, pcent, zero, turns, rate;

    if(args < 2)
    {
        fprintf(stderr, "Usage:\n%s [-lHz -rHz -ml%% -mr%%"\
                        " -tseconds] <filename>\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    twopi      = PI * 2.0;
    zero       = 0.0;
    rate       = (double)SAMPLE_RATE;
    left_freq  = LEFT_FREQ;
    right_freq = RIGHT_FREQ;
    sec        = SECONDS;
    dac_max    = 32767.5;
    pcent      = 100.0;
    left_mag   = dac_max;
    right_mag  = dac_max;
    fname      = nam;
    for(i=1; i < (size_t) args; i++)
    {
        if(argv[i][0] != (char) '-')
            fname = (const char *)argv[i];
        else
            switch((int)argv[i][1] & 95)
            {
            case 'L':
                left_freq = atof(&argv[i][2]);
                break;
            case 'M':
                if(argv[i][2] == (char) 'l')
                    left_mag = atof(&argv[i][3])  * dac_max / pcent;
                else if (argv[i][2] == (char) 'r')
                    right_mag = atof(&argv[i][3]) * dac_max / pcent;
                if(left_mag > dac_max)
                    left_mag = dac_max;
                else if(left_mag < zero)
                    left_mag = zero;
                if(right_mag > dac_max)
                    right_mag = dac_max;
                else if(right_mag < zero)
                    right_mag = zero;
                break;
            case 'R':
                right_freq = atof(&argv[i][2]);
                break;
            case 'T':
                sec = (size_t)atoi(&argv[i][2]);
                break;
            default:
                break;
            }
    }
    if(sec < 1) sec = 1;
    memcpy(wav.rif,  rif, sizeof(wav.rif));
    memcpy(wav.dat,  dat, sizeof(wav.dat));
    memcpy(wav.wav,  fmt, sizeof(wav.wav));
    wav.samp_rate  = U32(SAMPLE_RATE);
    wav.fmt_tag    = U16(WAV_FMT_PCM);
    wav.fmt_len    = U32(sizeof(wav.wav)
                   +     sizeof(wav.fmt_tag)
                   +     sizeof(wav.fmt_len)
                   +     sizeof(wav.chans));
    wav.chans      = U16(NR_CHANS);
    wav.block      = U16(NR_CHANS   * sizeof(int16_t));
    wav.bytes_sec  = U32(NR_CHANS   * sizeof(int16_t) * SAMPLE_RATE);
    len            = sec * NR_CHANS * sizeof(int16_t) * SAMPLE_RATE;
    wav.len        = U32(len - offsetof(WAV, dat));
    wav.dat_len    = U32(len);
    wav.bps        = U16(sizeof(int16_t) * CHAR_BITS);
    lim            = len / (NR_CHANS * sizeof(int16_t));
    printf("  Play length = %u seconds\n", sec);
    printf(" Left channel = %4.3f Hz (%3.3f%%)\n",
                  left_freq, left_mag * pcent / dac_max);
    printf("Right channel = %4.3f Hz (%3.3f%%)\n",
                  right_freq, right_mag * pcent /dac_max);
    if((fd = open(fname, O_RDWR|O_CREAT|O_TRUNC|O_BINARY, 0644)) == FAIL)
        ERRORS(open);
    if((buf = (int16_t *) malloc(len)) == NULL)
        ERRORS(malloc);
    if((size_t)write(fd, &wav, sizeof(wav)) != sizeof(wav))
        ERRORS(write);
    sp = buf;
    for(i= 0;  i < lim ; i++)
    {
        turns = ((double)i / rate) * twopi;
        theta = turns * left_freq;
        samp  = (int16_t) ((sin(theta) * left_mag) + FP_ROUND);
        *sp++ = I16(samp);                    // Left channel
        theta = turns * right_freq;
        samp  = (int16_t) ((sin(theta) * right_mag) + FP_ROUND);
        *sp++ = I16(samp);                    // Right channel
    }
    if((size_t)write(fd, buf, len) != len)
        ERRORS(write);
    free(buf);
    (void)close(fd);
    return 0;
}

