/*
 *  File: show_wave.c    Created 12-DEC-2001    Richard B. Johnson
 *                                              rjohnson@analogic.com
 *  Released under GPL.
 *  This displays the characteristics of a ".wav" file. In particular,
 *  it shows the maximum modulation that occurred with in the file.
 *  It can be used to set the recording levels.
 *
 *  It has been observed that some software converters, using /dev/dsp
 *  produce excessive distortion if the modulation level exceeds about
 *  85 percent. Even cheap CDROMs seem to take 100% modulation fine.
 *  This is incomplete. I only analyze PCM files, the most commonly
 *  used.
 */ 


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <sys/stat.h>


#ifdef _MSC_VER
#include "stdint.h"
#include <io.h>
#else
#include <unistd.h>
#include <stdint.h>
#include <endian.h>
#define O_BINARY 0
#endif

#include "wav.h"


typedef struct {
    WAV w;
    uint16_t d[2];
    } HDR;


HDR wav;
static const char rif[]="RIFF";
static const char typ[]="WAVEfmt ";


#define FAIL -1
#define DAC_MAX 0x7fff

#define ERRORS(s) \
{                 \
    fprintf(stderr, "Error on line %d, file %s, call "#s", (%s)\n", \
    __LINE__, __FILE__, strerror(errno));                \
    exit(EXIT_FAILURE);  \
}

int main(int args, char *argv[])
{
    int fd, status;
    size_t i, len, nr;
    char buf[0x100];
    char *fmt;
    int16_t *sp, *begin, sample, maxsamp, tmp;
    struct stat sb;

    if(args < 2)
    {
        fprintf(stderr, "Usage:\n%s <filename>\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    status = EXIT_SUCCESS;
    if((fd = open(argv[1], O_RDONLY)) == FAIL)
        ERRORS(open);
     if((size_t)read(fd, &wav, sizeof(wav)) != sizeof(wav))
        ERRORS(read);
     if(fstat(fd, &sb) == FAIL)
        ERRORS(fstat);
     printf("\tAnalysis of file %s\n", argv[1]);
     i  = memcmp(wav.w.rif, rif, sizeof(wav.w.rif)) |
          memcmp(wav.w.wav, typ, sizeof(wav.w.wav));
     printf("          File size = %ld\n", sb.st_size);
     memset(buf, 0x00, sizeof(buf));
     memcpy(buf, wav.w.rif, sizeof(wav.w.rif));
     printf("       Header ident = %s\n", buf);
     memset(buf, 0x00, sizeof(buf));
     memcpy(buf, wav.w.wav, sizeof(wav.w.wav));
     printf("       Format ident = %s\n", buf);
     printf("Length minus header = %u\n", U32(wav.w.len));
     printf("      Format length = %u\n", U32(wav.w.fmt_len));
     if(i)
     {
          fprintf(stderr, "Invalid file type!\n");
          exit(EXIT_FAILURE);
     }
     switch(U16(wav.w.fmt_tag))
     {
     case  WAV_FMT_MICROSLUT:
        fmt = "Microsoft";
        break;
     case WAV_FMT_PCM:
        fmt = "PCM";
        break;
     case WAV_FMT_ADPCM:
        fmt = "ADPCM";
        break;
     case WAV_FMT_IEEE_FLOAT:
        fmt = "IEEE-FLOAT";
        break;
     case WAV_FMT_ALAW:
        fmt = "ALAW";
        break;
     case WAV_FMT_MULAW:
        fmt = "MULAW";
     case WAV_FMT_OKI_ADPCM:
        fmt = "OKI-ADPCM";
        break;
     case WAV_FMT_IMA_ADPCM:
        fmt = "IMA-ADPCM";
        break;
     case WAV_FMT_DIGISTD:
        fmt = "DIGISTD";
        break;
     case WAV_FMT_DIGIFIX:
        fmt = "DIGIFIX";
        break;
     case WAV_FMT_DOLBY_AC2:
        fmt = "DOLBY-AC2";
        break;
     default:
        fmt = "unknown";
        break;
     }
     printf("             Format = %s\n", fmt);
     printf(" Number of channels = %d\n", (int)          U16(wav.w.chans));
     printf("        Sample rate = %u (Hz)\n",           U32(wav.w.samp_rate));
     printf("   Bytes per second = %u\n",                U32(wav.w.bytes_sec));
     printf("         Block size = %d (bytes)\n", (int)  U16(wav.w.block));
     printf("    Bits per sample = %d\n", (int)          U16(wav.w.bps));
     printf("        Data length = %u (bytes)\n",        U32(wav.w.dat_len));
     printf("         Data start = %04x %04x\n", I16(wav.d[0]), I16(wav.d[1]));
     printf("      Checking data ... ");
     fflush(stdout);

     len  = MIN((size_t)U32(wav.w.dat_len) - sizeof(wav.d), 0x100000);
     if((sp = (int16_t *) malloc(len)) == NULL)
         ERRORS(malloc);
	 begin = sp;
	 maxsamp = 0; 

     while((status = read(fd, sp, len)) > 0)
	 {
         nr = (size_t) status;
         nr /= sizeof(int16_t);
         for(i=0; i< nr; i++)
		{
            tmp    =  *sp++;
            sample = I16(tmp);
            if(sample < 0) sample = -sample;
            if(sample > maxsamp) maxsamp = sample;
            if(sample >= DAC_MAX)
			{
                printf("Overmodulation at %u\n", i);
                status = EXIT_FAILURE;
			}
		 }
	 }
     printf("done!\n");
     i = (size_t) maxsamp * 100 / DAC_MAX;
     printf("      Max amplitude = %d%%\n", i);
     free(begin);
     close(fd);
     return status;
}

