/*
 *  File: truncwave.c        Created 20-DEC-2001   Richard B. Johnson
 *                                                 rjohnson@analogic.com
 *
 *  Released under GPL.
 *
 *  This program removes starting and stopping dead-space from a ".wav"
 *  file that has been recorded. Typically, there is some time from the
 *  beginning of the recording of a session until the time at which a
 *  recorded performance begins. Also, when the performance ends, there
 *  is some elapsed time before the recording stops. If you put these
 *  files directly onto a CDROM, there will be a long pause between
 *  tracks. This reduces that long pause to more acceptable values.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include <stdint.h>
#include <endian.h>
#include "wav.h"

#define NoDEBUG

#ifdef DEBUG
#define DEB(f) f
#else
#define DEB(f)
#endif

#define FAIL -1
#define FUDGE   3.0 
#define ONE_SEC (44100 * sizeof(short) * 2)
#define BUF_LEN ONE_SEC
#define SCA_LEN (status / sizeof(short)) 
#define FDAC_MAX 32767.0
#define LSB (1.0 / FDAC_MAX)

#define ERRORS(s) \
{  \
   fprintf(stderr,"Error from line %d, file %s, %s (%s)\n", __LINE__,\
    __FILE__, s, strerror(errno)); \
   exit(EXIT_FAILURE); \
}
int main(int args, char *argv[])
{
    int in_fd, out_fd, status;
    char *buf;
    WAV *wp;
    int16_t *sp;
    size_t i, sec;
    double ave, noise;
    off_t start_pos, stop_pos, read_end;

    if(args < 3)
    {
        fprintf(stderr, "Usage:\n%s <infile> <outfile>\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    if((in_fd = open(argv[1], O_RDONLY, 0)) == FAIL)
        ERRORS("open");
    if((out_fd = open(argv[2], O_RDWR|O_CREAT|O_TRUNC, 0744)) == FAIL)
        ERRORS("open");
    if((buf = (char *) malloc(BUF_LEN * sizeof(char))) == NULL)
        ERRORS("malloc");
    wp = (WAV *) buf;
    if((size_t)read(in_fd, buf, sizeof(WAV)) != sizeof(WAV))
        ERRORS("read");
    if((size_t)write(out_fd, buf, sizeof(WAV)) != sizeof(WAV))
        ERRORS("write");
    fprintf(stdout, "Copy/truncate file %s to %s\n", argv[1], argv[2]);
/*
 *  Read the whole file and find the lowest average noise that exists
 *  within a one-second interval.
 */
    if((start_pos =  lseek(in_fd, 0, SEEK_CUR)) == FAIL)
        ERRORS("lseek");
    ave = 0.0;
    sec = 0;
    noise = FDAC_MAX;
    while((status = read(in_fd, buf, ONE_SEC)) > 0)
    {
        ave = 0.0;
        sp = (int16_t *) buf;
        for(i = 0; i< SCA_LEN; i++)
            ave += fabs((double) *sp++);
        ave /= (double) i;
        DEB(printf("Noise %f\n", ave));
        if((ave < noise) && (ave >= LSB))
            noise = ave;
    }
    DEB(printf("Average noise = %f\n", noise));
    noise *= FUDGE;
/*
 *  Check for no quiet areas at all.
 */
    if(noise > FDAC_MAX)
        noise = LSB * FUDGE;
/*
 *  Find out where audio starts and back up two seconds. We average
 *  because there may be some clicks that we don't want to think marks
 *  the beginning.
 */
    if(lseek(in_fd, start_pos, SEEK_SET) == FAIL)
        ERRORS("lseek");
    while((status = read(in_fd, buf, ONE_SEC)) > 0)
    {
        ave = 0.0;
        sp = (int16_t *) buf;
        for(i = 0; i< SCA_LEN; i++)
            ave += fabs((double) *sp++);
        ave /= (double) i;
        DEB(printf("Head %f\n", ave));
        if(ave > noise) break;
    }
    if((start_pos =  lseek(in_fd, 0, SEEK_CUR)) == FAIL)
        ERRORS("lseek");
    start_pos -= (ONE_SEC * 2);
    if(start_pos < (off_t)sizeof(WAV)) start_pos = (off_t)sizeof(WAV);
    fprintf(stdout, "Start offset = %ld ... ", start_pos);
    (void)fflush(stdout);
/*
 *  Find out where the audio stops. It will tail off so we adjust only
 *  to data boundaries.
 */
    while((status = read(in_fd, buf, ONE_SEC)) > 0)
    {
        sec++;
        ave = 0.0;
        sp = (int16_t *) buf;
        for(i = 0; i< SCA_LEN; i++)
            ave += fabs((double) *sp++);
        ave /= (double) i;
        DEB(printf("Tail %f\n", ave));
        if(ave <= noise)
            break;
    }
    if((read_end =  lseek(in_fd, 0, SEEK_CUR)) == FAIL)
        ERRORS("lseek");
    read_end += 3;
    read_end &= ~3; 
    fprintf(stdout, "Stop offset = %ld (%u) seconds\n", read_end, sec);
    stop_pos = read_end - (start_pos - (off_t) sizeof(WAV));
    if(lseek(in_fd, start_pos, SEEK_SET) == FAIL)
        ERRORS("lseek");
    while((status = read(in_fd, buf, BUF_LEN)) > 0)
    {
        if(write(out_fd, buf, (size_t)status) != status)
            ERRORS("write");
         if((start_pos += (off_t) status) >= read_end)
             break;
    }
    if(ftruncate(out_fd, stop_pos) == FAIL)
        ERRORS("ftruncate");
    if(lseek(out_fd, 0L, SEEK_SET) == FAIL)
        ERRORS("lseek");
    if((size_t)read(out_fd, buf, sizeof(WAV)) != sizeof(WAV))
        ERRORS("read");
    wp->dat_len = (uint32_t) stop_pos - sizeof(WAV);
    wp->len = wp->dat_len - (uint32_t) offsetof(WAV, dat);
    if(lseek(out_fd, 0L, SEEK_SET) == FAIL)
        ERRORS("lseek");
    if((size_t)write(out_fd, buf, sizeof(WAV)) != sizeof(WAV))
        ERRORS("write");
    (void)close(in_fd);
    (void)close(out_fd);
    free(buf);
    exit(EXIT_SUCCESS);
}
